use std::{marker::PhantomData, sync::Arc};

use crate::{
    flow::{
        binary::{balancing::*, bound::*, bounds::*, *},
        comparator::*,
        keyed::*,
    },
    func::context::*,
    rcore::*,
};

use super::*;

pub struct TreeContext<Cl, T>(Cl, PhantomData<T>);

impl<Cl: Clone, T> TreeContext<Cl, T> {
    pub fn new(cl: Cl) -> Self {
        Self(cl, PhantomData)
    }
}

impl<Cl: Clone, T> Clone for TreeContext<Cl, T> {
    fn clone(&self) -> Self {
        Self::new(self.0.clone())
    }
}

#[derive(Debug)]
pub enum TreeContextError<R, E> {
    Resolution(R),
    Extra(E),
}

pub type TreeContextError2<'a, Ctx, A, E> =
    TreeContextError<ResolutionFailure<'a, Ctx, Node<'a, Ctx, A>>, E>;

impl<L, P, E> From<ResolutionError<L, TreeParseError<P>>>
    for TreeContextError<ResolutionError<L, TreeParseError<P>>, E>
{
    fn from(value: ResolutionError<L, TreeParseError<P>>) -> Self {
        Self::Resolution(value)
    }
}

impl<L, P, E> From<TreeParseError<P>>
    for TreeContextError<ResolutionError<L, TreeParseError<P>>, E>
{
    fn from(value: TreeParseError<P>) -> Self {
        ResolutionError::Parse(value).into()
    }
}

impl<L, P, E> From<HeightError> for TreeContextError<ResolutionError<L, TreeParseError<P>>, E> {
    fn from(value: HeightError) -> Self {
        TreeParseError::HeightValue(value).into()
    }
}

pub type TreeContext2<'a, Ctx, A, C, E> = TreeContext<(Arc<C>, Fctr<'a, A>), (&'a (), Ctx, A, E)>;

impl<
        'a,
        Ctx: Context<'a>,
        A: Mentionable<'a, Ctx> + Clone,
        C: 'a + Comparator<A>,
        E: 'a + Send,
    > FunctorContext<'a> for TreeContext2<'a, Ctx, A, C, E>
{
    type T = FallibleMonad<'a, Ctx, TreeContextError2<'a, Ctx, A, E>>;
}

impl<
        'a,
        Ctx: Context<'a>,
        A: Mentionable<'a, Ctx> + Clone,
        C: 'a + Comparator<A>,
        E: 'a + Send,
    > Keyed for TreeContext2<'a, Ctx, A, C, E>
{
    type Key = A;
}

impl<
        'a,
        Ctx: Context<'a>,
        A: Mentionable<'a, Ctx> + Clone,
        C: 'a + Comparator<A>,
        E: 'a + Send,
    > WithComparator for TreeContext2<'a, Ctx, A, C, E>
{
    type Comparator = C;

    fn comparator(&self) -> &Self::Comparator {
        self.0 .0.as_ref()
    }
}

impl<
        'a,
        Ctx: Context<'a>,
        A: Mentionable<'a, Ctx> + Clone,
        C: 'a + Comparator<A>,
        E: 'a + Send,
    > BinaryTrees for TreeContext2<'a, Ctx, A, C, E>
{
    type Node = Node<'a, Ctx, A>;
    type Reference = Point<'a, Ctx, Self::Node>;
    type Tree = Tree<'a, Ctx, A>;

    fn split(&self, node: &Self::Node) -> crate::flow::binary::Split<Self> {
        (node.l.clone(), node.r.clone(), node.key.clone())
    }

    fn equal(&self, rl: &Self::Reference, rr: &Self::Reference) -> bool {
        rl == rr
    }

    fn refer(&self, tree: &Self::Tree) -> Option<Self::Reference> {
        tree.node.as_ref().cloned()
    }
}

impl<
        'a,
        Ctx: Context<'a>,
        A: Mentionable<'a, Ctx> + Clone,
        C: 'a + Comparator<A>,
        E: 'a + Send,
    > MonadTrees<'a> for TreeContext2<'a, Ctx, A, C, E>
{
    fn resolve(
        &self,
        reference: &Self::Reference,
    ) -> crate::flow::binary::BTWrap<'a, Self, Self::Node> {
        Ctx::stuff(reference.resolve_map(|resolved| {
            resolved
                .map(|rc| rc.as_ref().clone())
                .map_err(TreeContextError::Resolution)
        }))
    }
}

impl<
        'a,
        Ctx: Context<'a>,
        A: Mentionable<'a, Ctx> + Clone,
        C: 'a + Comparator<A>,
        E: 'a + Send,
    > BinaryTreesEmpty<'a> for TreeContext2<'a, Ctx, A, C, E>
{
    fn empty(&self) -> Self::Tree {
        Tree {
            node: Nullable::Null(NodeFactory {
                factory: self.0 .1.clone(),
                _ctx: PhantomData,
            }),
            height: 0,
        }
    }

    fn split_key_empty(
        &self,
        _tree: Self::Tree,
        _key: Self::Key,
    ) -> BTWrap<'a, Self, KeySplit<Self>> {
        Self::pure((self.empty(), self.empty()))
    }
}

impl<
        'a,
        Ctx: Context<'a>,
        A: Mentionable<'a, Ctx> + Clone,
        C: 'a + Comparator<A>,
        E: 'a + Send,
    > TreesHeight for TreeContext2<'a, Ctx, A, C, E>
{
    fn height(&self, tree: &Self::Tree) -> u64 {
        tree.height
    }
}

impl<
        'a,
        Ctx: Context<'a>,
        A: Mentionable<'a, Ctx> + Clone,
        C: 'a + Comparator<A>,
        E: 'a + Send,
    > BinaryTreesTryJoin<'a> for TreeContext2<'a, Ctx, A, C, E>
{
    fn try_join(
        &self,
        tl: Self::Tree,
        key: Self::Key,
        tr: Self::Tree,
    ) -> BTWrap<'a, Self, Self::Node> {
        Self::pure(Node { l: tl, r: tr, key })
    }
}

impl<
        'a,
        Ctx: Context<'a>,
        A: Mentionable<'a, Ctx> + Clone,
        C: 'a + Comparator<A>,
        E: 'a + Send,
    > TreesHeightError<'a> for TreeContext2<'a, Ctx, A, C, E>
{
    fn height_error<T: 'a + Send>(&self, error: HeightError) -> BTWrap<'a, Self, T> {
        Self::fail(error.into())
    }
}

// #[derive(Debug)]
// pub enum WrappedExtra<W, E> {
//     Wrapped(W),
//     Extra(E),
// }

// impl<R, E, W> From<WrappedExtra<W, E>> for TreeContextError<R, WrappedExtra<W, E>> {
//     fn from(value: WrappedExtra<W, E>) -> Self {
//         Self::Extra(value)
//     }
// }

// impl<R, E> From<BalancingError> for TreeContextError<R, WrappedExtra<BalancingError, E>> {
//     fn from(value: BalancingError) -> Self {
//         WrappedExtra::Wrapped(value).into()
//     }
// }

// impl<R, A, E> From<BoundsError<A>> for TreeContextError<R, WrappedExtra<BoundsError<A>, E>> {
//     fn from(value: BoundsError<A>) -> Self {
//         WrappedExtra::Wrapped(value).into()
//     }
// }

impl<
        'a,
        Ctx: Context<'a>,
        A: Mentionable<'a, Ctx> + Clone,
        C: 'a + Comparator<A>,
        E: 'a + Send + From<BalancingError>,
    > BinaryTreesUnbalanced<'a> for TreeContext2<'a, Ctx, A, C, E>
{
    fn tree_of_with_height(&self, node: Self::Node, height: u64) -> BTWrap<'a, Self, Self::Tree> {
        let node = Nullable::from(node);
        Self::pure(Tree { node, height })
    }

    fn balancing_error<T: 'a + Send>(&self, error: BalancingError) -> BTWrap<'a, Self, T> {
        Self::fail(TreeContextError::Extra(error.into()))
    }

    fn node_heights(&self, node: &Self::Node) -> (u64, u64) {
        (node.l.height, node.r.height)
    }
}

impl<
        'a,
        Ctx: Context<'a>,
        A: Mentionable<'a, Ctx> + Clone,
        C: 'a + Comparator<A>,
        E: 'a + Send + From<BoundsError<A>>,
    > BinaryTreesBindable<'a> for TreeContext2<'a, Ctx, A, C, E>
{
    fn bounds_error<T: 'a + Send>(&self, error: BoundsError<Self::Key>) -> BTWrap<'a, Self, T> {
        Self::fail(TreeContextError::Extra(error.into()))
    }
}

#[derive(Debug)]
pub enum BbError<A> {
    Balancing(BalancingError),
    Bounds(BoundsError<A>),
}

impl<A> From<BalancingError> for BbError<A> {
    fn from(value: BalancingError) -> Self {
        Self::Balancing(value)
    }
}

impl<A> From<BoundsError<A>> for BbError<A> {
    fn from(value: BoundsError<A>) -> Self {
        Self::Bounds(value)
    }
}
